# Copyright 2021 MosaicML. All Rights Reserved.

from composer.models.classify_mnist.mnist_hparams import MnistClassifierHparams as MnistClassifierHparams
from composer.models.classify_mnist.model import MNIST_Classifier as MNIST_Classifier

_task = 'Image Classification'
_dataset = 'MNIST'
_name = 'SimpleConvNet'
_quality = ''
_metric = 'Accuracy'
_ttt = '?'
_hparams = 'hparams_cpu.yaml'
