# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.colout.colout import ColOut as ColOut
from composer.algorithms.colout.colout import ColOutHparams as ColOutHparams
from composer.algorithms.colout.colout import colout as colout

_name = 'ColumnOut'
_class_name = 'ColOut'
_functional = 'colout'
_tldr = 'Removes columns and rows from the image for augmentation and efficiency.'
_attribution = 'MosaicML'
_link = ''
_method_card = ''
