# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.scale_schedule.scale_schedule import ScaleSchedule as ScaleSchedule
from composer.algorithms.scale_schedule.scale_schedule import ScaleScheduleHparams as ScaleScheduleHparams
from composer.algorithms.scale_schedule.scale_schedule import scale_scheduler as scale_scheduler

_name = 'Scale Schedule'
_class_name = 'ScaleScheduler'
_functional = 'scale_scheduler'
_tldr = 'Scales the learning schedule by a factor'
_attribution = ''
_link = ''
_method_card = ''
