# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.sam.sam import SAM as SAM
from composer.algorithms.sam.sam import SAMHparams as SAMHparams

_name = 'SAM'
_class_name = 'SAM'
_functional = ''
_tldr = 'SAM optimizer measures sharpness of optimization space'
_attribution = '(Foret et al, 2021)'
_link = 'https://arxiv.org/abs/2010.01412'
_method_card = ''
