# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.no_op_model.no_op_model import NoOpModel as NoOpModel
from composer.algorithms.no_op_model.no_op_model import NoOpModelHparams as NoOpModelHparams

_name = 'No-Op Model'
_class_name = 'NoOpModel'
_functional = ''
_tldr = 'Replaces the model with a dummy model. Used for dataloader profiling.'
_attribution = 'MosaicML'
_link = ''
_method_card = ''
