# Copyright 2021 MosaicML. All Rights Reserved.

from composer.algorithms.channels_last.channels_last import ChannelsLast as ChannelsLast
from composer.algorithms.channels_last.channels_last import ChannelsLastHparams as ChannelsLastHparams

_name = 'Channels Last'
_class_name = 'ChannelsLast'
_functional = ''
_tldr = 'Uses channels last memory format (NHWC)'
_attribution = 'PyTorch'
_link = 'https://pytorch.org/tutorials/intermediate/memory_format_tutorial.html'
_method_card = ''
